/*
    Copyright (C) 2010 wpitchoune@gmail.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
    02110-1301 USA 
*/

#include <stdlib.h>
#include <string.h>

#include <sensors/sensors.h>
#include <sensors/error.h>

#include "config.h"
#include "hdd.h"
#include "psensor.h"
#include "nvidia.h"
#include "lmsensor.h"

struct psensor *psensor_create(char *id, char *name, int type)
{
	struct psensor *psensor
	    = (struct psensor *)malloc(sizeof(struct psensor));
	int t;

	psensor->id = id;
	psensor->name = name;
	psensor->enabled = 1;
	psensor->min = 0;
	psensor->max = 0;

	psensor->type = type;

	for (t = 0; t < MAX_MEASURES; t++)
		psensor->values[t] = 0;

	psensor->alarm_limit = 0;

	psensor->cb_alarm_raised = NULL;
	psensor->cb_alarm_raised_data = NULL;
	psensor->alarm_raised = 0;

	psensor->alarm_enabled = 0;

	return psensor;
}

void psensor_free(struct psensor *sensor)
{
	if (sensor) {
		free(sensor->name);
		free(sensor->id);

		if (sensor->color)
			free(sensor->color);

		free(sensor);
	}
}

void psensor_list_free(struct psensor **sensors)
{
	struct psensor **sensor_cur;

	if (sensors) {
		sensor_cur = sensors;

		while (*sensor_cur) {
			psensor_free(*sensor_cur);

			sensor_cur++;
		}

		free(sensors);

		sensors = NULL;
	}
}

int psensor_list_size(struct psensor **sensors)
{
	int size;
	struct psensor **sensor_cur;

	if (!sensors)
		return 0;

	size = 0;
	sensor_cur = sensors;

	while (*sensor_cur) {
		size++;
		sensor_cur++;
	}
	return size;
}

int psensor_list_contains_type(struct psensor **sensors, int type)
{
	struct psensor **s;

	if (!sensors)
		return 0;

	s = sensors;
	while (*s) {
		if ((*s)->type == type)
			return 1;
		s++;
	}

	return 0;
}

struct psensor **psensor_list_add(struct psensor **sensors,
				  struct psensor *sensor)
{
	int size = psensor_list_size(sensors);

	struct psensor **result
	    = malloc((size + 1 + 1) * sizeof(struct psensor *));

	if (sensors) {
		memcpy(result, sensors, size * sizeof(struct psensor *));
	}

	result[size] = sensor;
	result[size + 1] = NULL;

	return result;
}

int is_temp_type(int type)
{
	return type & SENSOR_TYPE_TEMP;
}

int is_fan_type(int type)
{
	return type & SENSOR_TYPE_FAN;
}

char *psensor_value_to_string(int type, double value)
{
	/* should not be possible to exceed 20 characters with temp or
	   rpm values the .x part is never displayed */
	char *str = malloc(20);

	char *unit;

	if (is_temp_type(type))
		unit = "C";
	else
		unit = "";

	sprintf(str, "%.0f%s", value, unit);

	return str;
}

void psensor_set_current_value(struct psensor *sensor, double value)
{
	memmove(sensor->values,
		&sensor->values[1], (MAX_MEASURES - 1) * sizeof(double));

	sensor->values[MAX_MEASURES - 1] = value;

	if (!sensor->min || value < sensor->min)
		sensor->min = value;

	if (!sensor->max || value > sensor->max)
		sensor->max = value;

	if (sensor->alarm_limit && sensor->alarm_enabled) {

		if (value > sensor->alarm_limit) {

			if (!sensor->alarm_raised && sensor->cb_alarm_raised) {
				sensor->cb_alarm_raised
				    (sensor, sensor->cb_alarm_raised_data);
			}

			sensor->alarm_raised = 1;

		} else {
			sensor->alarm_raised = 0;
		}

	}
}

double psensor_get_current_value(struct psensor *sensor)
{
	return sensor->values[MAX_MEASURES - 1];
}

/*
  Returns the minimal value of a given 'type' (SENSOR_TYPE_TEMP or
  SENSOR_TYPE_FAN)
 */
double get_min_value(struct psensor **sensors, int type)
{
	int m = 0;
	struct psensor **s = sensors;

	while (*s) {
		struct psensor *sensor = *s;

		if (sensor->enabled && (sensor->type & type)) {
			int i;
			int t;
			for (i = 0; i < MAX_MEASURES; i++) {
				t = sensor->values[i];

				if (t && (t < m || m == 0))
					m = t;
			}
		}
		s++;
	}

	return m;

}

/*
  Returns the maximal value of a given 'type' (SENSOR_TYPE_TEMP or
  SENSOR_TYPE_FAN)
 */
double get_max_value(struct psensor **sensors, int type)
{
	int m = 0;
	struct psensor **s = sensors;

	while (*s) {
		struct psensor *sensor = *s;

		if (sensor->enabled && (sensor->type & type)) {
			int i;
			int t;
			for (i = 0; i < MAX_MEASURES; i++) {
				t = sensor->values[i];

				if (t > m)
					m = t;
			}
		}
		s++;
	}

	return m;

}

int get_min_temp(struct psensor **sensors)
{
	return get_min_value(sensors, SENSOR_TYPE_TEMP);
}

int get_min_rpm(struct psensor **sensors)
{
	return get_min_value(sensors, SENSOR_TYPE_FAN);
}

int get_max_rpm(struct psensor **sensors)
{
	return get_max_value(sensors, SENSOR_TYPE_FAN);
}

int get_max_temp(struct psensor **sensors)
{
	return get_max_value(sensors, SENSOR_TYPE_TEMP);
}

struct psensor **get_all_sensors()
{
	struct psensor **psensors = NULL;
	int count = 0;
	const sensors_chip_name *chip;
	int chip_nr = 0;
	struct psensor **tmp_psensors;
#ifdef NVIDIA_SUPPORT
	int i, nvidia_gpus_count;
#endif

	while ((chip = sensors_get_detected_chips(NULL, &chip_nr))) {
		const sensors_feature *feature;
		int i = 0;
		while ((feature = sensors_get_features(chip, &i))) {
			struct psensor *psensor;

			if ((feature->type == SENSORS_FEATURE_TEMP
			     || feature->type == SENSORS_FEATURE_FAN)
			    && (psensor = lmsensor_psensor_create(chip,
								  feature))) {
				struct psensor **tmp_psensors
				    = psensor_list_add(psensors,
						       psensor);

				free(psensors);

				psensors = tmp_psensors;

				count++;
			}
		}
	}

#ifdef NVIDIA_SUPPORT
	nvidia_gpus_count = nvidia_init();

	if (!nvidia_gpus_count) {
		fprintf(stderr,
			"ERROR: no nvidia chips or initialization failure\n");
	}

	for (i = 0; i < nvidia_gpus_count; i++) {
		struct psensor *sensor = nvidia_create_sensor(i);

		struct psensor **tmp_psensors = psensor_list_add(psensors,
								 sensor);

		free(psensors);

		psensors = tmp_psensors;

		count++;
	}
#endif

	tmp_psensors = hdd_psensor_list_add(psensors);

	if (tmp_psensors != psensors) {
		free(psensors);
		psensors = tmp_psensors;
	}

	if (!psensors) {
		psensors = malloc(sizeof(struct psensor *));
		*psensors = NULL;
	}

	return psensors;
}

const char *psensor_type_to_str(int type)
{
	switch (type) {
	case SENSOR_TYPE_LMSENSOR_TEMP:
		return "Temperature";
	case SENSOR_TYPE_LMSENSOR_FAN:
		return "Fan";
	case SENSOR_TYPE_NVIDIA:
		return "NVidia GPU Temperature";
	case SENSOR_TYPE_HDD_TEMP:
		return "HDD Temperature";
	default:
		return "N/A";
	}
}
